import json
import os
import tempfile
import copy

from cba.utils import agent_utils as au
from cba import prompts
from langchain.schema import HumanMessage, SystemMessage
from cba.utils.metrics_utils import score_insight
from cba import metrics
from PIL import Image


class Agent:
    def __init__(
        self,
        df,
        df_user,
        gen_engine,
        dataset_csv_path=None,
        user_dataset_csv_path=None,
        goal=None,
        max_questions=3,
        branch_depth=4,
        n_retries=2,
        savedir=None,
    ):
        self.df = df
        self.df_user = df_user
        self.goal = goal
        self.max_questions = max_questions

        self.gen_engine = gen_engine
        self.n_retries = n_retries
        self.branch_depth = branch_depth

        if savedir is None:
            savedir = tempfile.mkdtemp()
        self.savedir = savedir
        if dataset_csv_path is None:
            dataset_csv_path = os.path.join(savedir, "dataset.csv")
            df.to_csv(dataset_csv_path, index=False)

        self.dataset_csv_path = dataset_csv_path
        self.user_dataset_csv_path = user_dataset_csv_path
        self.agent_poirot = AgentPoirot(
            df=df,
            df_user=df_user,
            gen_engine=gen_engine,
            savedir=savedir,
            goal=goal,
            verbose=True,
            dataset_csv_path=dataset_csv_path,
            user_dataset_csv_path=user_dataset_csv_path,
        )

    def run_agent(self) -> tuple:
        """
        run the agent to generate a sequence of questions and answers
        """
        # Prompt 2: Get Root Questions
        root_questions = self.agent_poirot.recommend_questions(
            prompt_method="basic", n_questions=self.max_questions
        )

        # Go through the root questions and generate insights
        for q in root_questions:
            question = q
            for i in range(self.branch_depth):
                if self.df_user is None:
                    prompt_code_method = "single"
                else:
                    prompt_code_method = "multi"
                answer, insight_dict = self.agent_poirot.answer_question(
                    question,
                    prompt_code_method=prompt_code_method,
                    prompt_interpret_method="basic",
                )

                next_questions = self.agent_poirot.recommend_questions(
                    n_questions=self.max_questions,
                    insights_history=[insight_dict],
                    prompt_method="follow_up_with_type",
                    question_type="descriptive",
                )
                question = next_questions[
                    self.agent_poirot.select_a_question(next_questions)
                ]

        self.agent_poirot.save_state_dict(
            os.path.join(self.savedir, "insights_history.json")
        )

        return self.agent_poirot.insights_history

    def load_checkpoint(self, savedir):
        self.agent_poirot.load_state_dict(
            os.path.join(savedir, "insights_history.json")
        )

    def evaluate_agent_on_summary(
        self, gt_insights_dict, score_name, return_summary=False
    ):
        # Get Summary Evaluation
        pred_summary = self.agent_poirot.summarize()
        gt_summary = gt_insights_dict["flag"]
        score_summary = score_insight(pred_summary, gt_summary, score_name=score_name)

        if return_summary:
            summary_dict = {
                "score_summary": score_summary,
                "pred_summary": pred_summary,
                "gt_summary": gt_summary,
            }
            return score_summary, summary_dict
        return score_summary

    def evaluate_agent_on_notebook(self, gt_flags_dict, score_method="rouge1"):
        """
        Evaluate the agent's performance
        """
        # get groundtruth
        gt_insights = []
        for o in gt_flags_dict["insights"]:
            gt_insights += [o["insight_dict"]["insight"]]

        pred_insights = [o["answer"] for o in self.agent_poirot.insights_history]
        # compute score using score_method
        if score_method == "rouge1":
            return metrics.compute_rouge(pred_insights, gt_insights)
        elif score_method == "g_eval":
            return metrics.compute_g_eval_m2m(pred_insights, gt_insights)
        elif score_method == "llama3_eval":
            return metrics.compute_llama3_eval_m2m(pred_insights, gt_insights)


class AgentPoirot:
    def __init__(
        self,
        df,
        df_user,
        savedir=None,
        context="This is a dataset that could potentially consist of interesting insights",
        gen_engine="gpt-3.5-turbo-0613",
        goal="I want to find interesting trends in this dataset",
        verbose=False,
        dataset_csv_path=None,
        user_dataset_csv_path=None,
    ):
        self.goal = goal
        self.df = df
        if savedir is None:
            savedir = tempfile.mkdtemp()
        self.savedir = savedir
        self.context = context
        # save df without index in savedir
        if dataset_csv_path is None:
            self.dataset_csv_path = os.path.join(savedir, "dataset.csv")
            self.df.to_csv(self.dataset_csv_path, index=False)
        else:
            self.dataset_csv_path = dataset_csv_path
            self.user_dataset_csv_path = user_dataset_csv_path
        self.gen_engine = gen_engine
        self.schema = au.get_schema(df)
        if df_user is not None:
            self.user_schema = au.get_schema(df_user)
        else:
            self.user_schema = None
        self.insights_history = []
        self.verbose = verbose

    def summarize(self, method="list", prompt_summarize_method="basic"):
        if method == "list":
            chat = au.get_chat_model(self.gen_engine)

            # Function to format the data
            def format_data(data):
                result = ""
                for i, item in enumerate(data):
                    question_tag = f"<question_{i}>{item['question']}</question_{i}>\n"
                    answer_tag = f"<answer_{i}>{item['answer']}</answer_{i}>\n\n"
                    result += f"{question_tag} {answer_tag}\n"
                return result

            # Format the data and print
            formatted_history = format_data(self.insights_history)

            # summary = agent.summarize_insights(method="list")
            system_prompt, content_prompt = prompts.get_summarize_prompt(
                method=prompt_summarize_method
            )
            messages = [
                SystemMessage(content=system_prompt),
                HumanMessage(
                    content=content_prompt.format(
                        context=self.context,
                        goal=self.goal,
                        history=formatted_history,
                    )
                ),
            ]

            def _validate_tasks(out):
                isights = au.extract_html_tags(out, ["insight"])

                # Check that there are insights generated
                if "insight" not in isights:
                    return (
                        out,
                        False,
                        f"Error: you did not generate insights within the <insight></insight> tags.",
                    )
                isights = isights["insight"]
                return (isights, out), True, ""

            insight_list, message = au.chat_and_retry(
                chat, messages, n_retry=3, parser=_validate_tasks
            )

            insights = "\n".join(insight_list)

        return insights

    def select_a_question(self, questions):
        """
        Select a question from the list of questions
        """
        return au.select_a_question(
            questions,
            self.context,
            self.goal,
            [o["question"] for o in self.insights_history],
            self.gen_engine,
            prompts.SELECT_A_QUESTION_TEMPLATE,
            prompts.SELECT_A_QUESTION_SYSTEM_MESSAGE,
        )

    def generate_notebook():
        pass

    def generate_report():
        pass

    def recommend_questions(
        self,
        n_questions=3,
        insights_history=None,
        prompt_method=None,
        question_type=None,
    ):
        """
        Suggest Next Best Questions
        """
        if self.verbose:
            print(f"Generating {n_questions} Questions using {self.gen_engine}...")

        if insights_history is None:

            # Generate Root Questions
            questions = au.get_questions(
                prompt_method=prompt_method,
                context=self.context,
                goal=self.goal,
                messages=[],
                schema=self.schema,
                max_questions=n_questions,
                model_name=self.gen_engine,
            )
        else:
            # Generate Follow Up Questions
            last_insight = insights_history[-1]
            questions = au.get_follow_up_questions(
                context=self.context,
                goal=self.goal,
                question=last_insight["question"],
                answer=last_insight["answer"],
                schema=self.schema,
                max_questions=n_questions,
                model_name=self.gen_engine,
                prompt_method=prompt_method,
                question_type=question_type,
            )
            if self.verbose:
                print(
                    "\nFollowing up on the last insight:\n---------------------------------"
                )
                print(f"Question: {last_insight['question']}\n")
                print(f"Answer: {last_insight['answer']}\n")

        if self.verbose:
            print("\nNext Best Questions:\n-------------------")
            for idx, question in enumerate(questions):
                print(f"{idx+1}. {question}")
            print()

        return questions

    def answer_question(
        self,
        question,
        n_retries=2,
        return_insight_dict=True,
        prompt_code_method="single",
        prompt_interpret_method="interpret",
    ):
        if self.verbose:
            print(f"Generating Code...")
        # Prompt 3: Generate Code
        code_output_folder = os.path.join(
            self.savedir, f"question_{str(len(self.insights_history))}"
        )

        if self.verbose:
            print(f"Interpreting Solution...")
            print(f"Results saved at: {self.savedir}")

        # with au.SuppressOutput():
        solution = au.generate_code(
            schema=self.schema,
            user_schema=self.user_schema,
            goal=self.goal,
            question=question,
            database_path=os.path.abspath(self.dataset_csv_path),
            user_database_path=(
                os.path.abspath(self.user_dataset_csv_path)
                if self.user_dataset_csv_path is not None
                else None
            ),
            output_folder=code_output_folder,
            model_name=self.gen_engine,
            n_retries=n_retries,
            prompt_method=prompt_code_method,
        )

        # Prompt 4: Interpret Solution
        interpretation_dict = au.interpret_solution(
            solution=solution,
            model_name=self.gen_engine,
            n_retries=n_retries,
            prompt_method=prompt_interpret_method,
        )
        answer = interpretation_dict["interpretation"]["answer"]

        if self.verbose:
            print("\nSolution\n---------")
            print(f"Question: {question}\n")
            print(f"Answer: {answer}\n")
            print(
                f"Justification: {interpretation_dict['interpretation']['justification']}\n"
            )

        insight_dict = {
            "question": question,
            "answer": answer,
            "justification": interpretation_dict["interpretation"]["justification"],
            "output_folder": code_output_folder,
        }

        # Save into the savedir
        with open(os.path.join(code_output_folder, "insight.json"), "w") as json_file:
            json.dump(insight_dict, json_file, indent=4, sort_keys=True)

        # add to insights
        self.insights_history += [insight_dict]

        insight_dict = copy.deepcopy(insight_dict)
        insight_dict.update(self.get_insight_objects(insight_dict))

        if return_insight_dict:
            return answer, insight_dict

        return answer["answer"]

    def get_insight_objects(self, insight_dict):
        """
        Get Insight Objects
        """
        if os.path.exists(os.path.join(insight_dict["output_folder"], "plot.jpg")):
            # get plot.jpg
            plot = Image.open(os.path.join(insight_dict["output_folder"], "plot.jpg"))
        else:
            plot = None

        if os.path.exists(os.path.join(insight_dict["output_folder"], "x_axis.jpg")):
            # get x_axis.json
            x_axis = json.load(
                open(os.path.join(insight_dict["output_folder"], "x_axis.json"), "r")
            )
        else:
            x_axis = None

        if os.path.exists(os.path.join(insight_dict["output_folder"], "y_axis.json")):
            # get y_axis.json
            y_axis = json.load(
                open(os.path.join(insight_dict["output_folder"], "y_axis.json"), "r")
            )
        else:
            y_axis = None

        if os.path.exists(os.path.join(insight_dict["output_folder"], "stat.json")):
            # get stat.json
            stat = json.load(
                open(os.path.join(insight_dict["output_folder"], "stat.json"), "r")
            )
        else:
            stat = None

        # get code.py
        if os.path.exists(os.path.join(insight_dict["output_folder"], "code.py")):
            code = open(
                os.path.join(insight_dict["output_folder"], "code.py"), "r"
            ).read()
        else:
            code = None

        insight_object = {
            "plot": plot,
            "x_axis": x_axis,
            "y_axis": y_axis,
            "stat": stat,
            "code": code,
        }
        return insight_object

    def save_state_dict(self, fname):
        with open(fname, "w") as f:
            json.dump(self.insights_history, f, indent=4)

    def load_state_dict(self, fname):
        with open(fname, "r") as f:
            self.insights_history = json.load(f)
