"""
Tests for the naive code generator.

"""

import pytest, os
import tempfile

# We should mostly test the code parser

from cba.utils.agent_utils import _code_parser


code_exception = [
    ("prin('Hello World)", "SyntaxError"),
    ("prin('Hello World')", "NameError"),
]


def test_detects_code_exceptions():
    """
    Test that the code parser detects with exceptions occur at runtime.

    """

    for code, exception in code_exception:
        with tempfile.TemporaryDirectory() as OUTPUT_FOLDER:
            os.makedirs(OUTPUT_FOLDER, exist_ok=True)
            out, valid, error_msg = _code_parser(code, OUTPUT_FOLDER)
            assert not valid
            assert exception.lower() in error_msg.lower()


def test_executes_code_correctly():
    """
    Test that the code parser executes code correctly.

    """
    with tempfile.TemporaryDirectory() as OUTPUT_FOLDER:
        os.makedirs(OUTPUT_FOLDER, exist_ok=True)
        files_to_create = ["x_axis.json", "y_axis.json", "plot.jpg", "stat.json"]
        code = f"print('Hello World')\n[open(f, 'w') for f in '{','.join(files_to_create)}'.split(',')]"
        out, valid, error_msg = _code_parser(code, OUTPUT_FOLDER)
        assert valid
        assert error_msg is ""
        assert out.strip() == "Hello World"


def test_validate_file_names():
    """
    Test that the parser detects errors in json files

    """
    for filename, is_valid in [
        # Json files
        ("x_axis.json", True),
        ("y_axis.json", True),
        ("xaxis.json", False),
        ("xaxis_.json", False),
        ("xaxis_a.json", False),
        ("xaxis1.json", False),
        ("yaxis.json", False),
        ("yaxis_.json", False),
        ("yaxis_a.json", False),
        ("yaxis1.json", False),
        ("z_axis_1.json", False),
        ("sta.json", False),
        ("stat.json", True),
        ("stat_.json", False),
        ("stat_a.json", False),
        ("x_axis_1.json", False),
        ("y_axis_1.json", False),
        ("stat_1.json", False),
        # Plot files
        ("plot.jpg", True),
        ("plot_.jpg", False),
        ("plot_a.jpg", False),
        ("plot1.jpg", False),
        ("plot_1.jpg", False),
    ]:
        with tempfile.TemporaryDirectory() as OUTPUT_FOLDER:
            os.makedirs(OUTPUT_FOLDER, exist_ok=True)
            # Create some files that are necessary for the test not to fail
            # For example, the parser will say that x_axis_1.json is not valid
            # on its own. You need to have y_axis_1.json too.
            files_to_create = ["x_axis.json", "y_axis.json", "plot.jpg", "stat.json"]

            # Create the file and check if it is valid
            files_to_create += [filename]
            code = f"[open(f, 'w') for f in '{','.join(files_to_create)}'.split(',')]"
            _, valid, _ = _code_parser(code, OUTPUT_FOLDER)
            assert valid == is_valid


if __name__ == "__main__":
    test_detects_code_exceptions()
    test_executes_code_correctly()
    test_validate_file_names()
